;;; doc-view.el --- View PDF/PostScript/DVI files in Emacs

;; Copyright (C) 2007 Free Software Foundation, Inc.
;;
;; Author: Tassilo Horn <tassilo@member.fsf.org>
;; Maintainer: Tassilo Horn <tassilo@member.fsf.org>
;; Keywords: files, pdf, ps, dvi
;; Version: <2007-08-25 Sat 15:06>

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.

;; GNU Emacs is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.

;; You should have received a copy of the GNU General Public License along with
;; GNU Emacs; see the file COPYING.  If not, write to the Free Software
;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
;; USA.

;;; Requirements:

;; I tested in on GNU Emacs 22, but maybe it works with older emacsen or
;; XEmacs, too.  You need GhostScript and `dvipdfm' which comes with TeTeX.

;;; Commentary:

;; DocView is a document viewer for Emacs.  It converts PDF, PS and DVI files
;; to a set of PNG files, one PNG for each page, and displays the PNG images
;; inside an Emacs buffer.  This buffer uses `doc-view-mode' which provides
;; convenient key bindings for browsing the document.
;;
;; To use it simply do
;;
;;     M-x doc-view RET
;;
;; and you'll be queried for a document to open.
;;
;; Since conversion may take some time all the PNG images are cached in a
;; subdirectory of `doc-view-cache-directory' and reused when you want to view
;; that file again.  This reusing can be omitted if you provide a prefx
;; argument to `doc-view'.  To delete all cached files use
;; `doc-view-clear-cache'.  To open the cache with dired, so that you can tidy
;; it out use `doc-view-dired-cache'.

;;; Code:

(require 'dired)

(defgroup doc-view nil
  "In-buffer viewer for PDF, PostScript and DVI files."
  :link '(function-link doc-view)
  ;; :version "22.2"
  :group 'applications
  :group 'multimedia
  :prefix "doc-view-")

(defcustom doc-view-ghostscript-program "gs"
  "Program to convert PS and PDF files to PNG."
  :type '(file)
  :group 'doc-view)

(defcustom doc-view-ghostscript-options '("-dNOPAUSE" "-sDEVICE=png16m" "-dTextAlphaBits=4"
					  "-dBATCH" "-dGraphicsAlphaBits=4" "-dQUIET"
					  "-r100")
  "A list of options to give to ghostview."
  :type '(sexp)
  :group 'doc-view)

(defcustom doc-view-dvipdfm-program "dvipdfm"
  "Program to convert DVI files to PDF.

DVI file will be converted to PDF before the resulting PDF is
converted to PNG."
  :type '(file)
  :group 'doc-view)

(defcustom doc-view-cache-directory (concat temporary-file-directory "doc-view")
  "The base directory, where the PNG images will be saved."
  :type '(directory)
  :group 'doc-view)

(defcustom doc-view-display-margin 5
  "The width of the margin put around the page's image."
  :type '(integer)
  :group 'doc-view)

(defvar doc-view-current-files nil
  "Only used internally.")

(defvar doc-view-current-page nil
  "Only used internally.")

(defvar doc-view-current-doc nil
  "Only used internally.")

(defvar doc-view-current-converter-process nil
  "Only used internally.")

(defvar doc-view-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map (kbd "C-v")       'doc-view-next-page)
    (define-key map (kbd "n")         'doc-view-next-page)
    (define-key map (kbd "<mouse-5>") 'doc-view-next-page)
    (define-key map (kbd "SPC")       'doc-view-next-page)
    (define-key map (kbd "M-v")       'doc-view-previous-page)
    (define-key map (kbd "p")         'doc-view-previous-page)
    (define-key map (kbd "<mouse-4>") 'doc-view-previous-page)
    (define-key map (kbd "M-<")       'doc-view-first-page)
    (define-key map (kbd "M->")       'doc-view-last-page)
    (define-key map (kbd "g")         'doc-view-goto-page)
    (define-key map (kbd "k")         'doc-view-kill-proc-and-buffer)
    (define-key map (kbd "C-x k")     'doc-view-kill-proc-and-buffer)
    (define-key map (kbd "q")         'bury-buffer)
    (suppress-keymap map)
    map)
  "Keymap used by `doc-view-mode'.")

(defun doc-view-kill-proc-and-buffer ()
  "Kill the current converter process and buffer."
  (interactive)
  (when (eq major-mode 'doc-view-mode)
    (when doc-view-current-converter-process
      (kill-process doc-view-current-converter-process))
    (kill-buffer (current-buffer))))

(defun doc-view-goto-page (page)
  "View the page given by ARG."
  (interactive "nPage: ")
  (if (< page 1)
      (setq page 1)
    (when (> page (length doc-view-current-files))
      (setq page (length doc-view-current-files))))
  (setq doc-view-current-page page)
  (doc-view-display-image (nth (1- page) doc-view-current-files)))

(defun doc-view-next-page (arg)
  "Browse ARG pages forward."
  (interactive "p")
  (doc-view-goto-page (+ doc-view-current-page arg)))

(defun doc-view-previous-page (arg)
  "Browse ARG pages backward."
  (interactive "p")
  (doc-view-goto-page (- doc-view-current-page arg)))

(defun doc-view-first-page ()
  "View the first page."
  (interactive)
  (doc-view-goto-page 1))

(defun doc-view-last-page ()
  "View the last page."
  (interactive)
  (doc-view-goto-page (length doc-view-current-files)))

(defun doc-view-file-name-to-directory-name (file)
  "Return the directory where the png files of FILE should be saved.

It'a a subdirectory of `doc-view-cache-directory'."
  (concat (file-name-as-directory doc-view-cache-directory)
	  (with-temp-buffer
	    (insert-file-contents-literally file)
	    (md5 (current-buffer)))))

(defun doc-view-convert-doc (doc)
  "Convert DOC to a set of png files, one file per page.

Those files are saved in the directory given by
`doc-view-file-name-to-directory-name'.

Returns the converter process."
  (clear-image-cache)
  (let* ((real-doc doc)
	 (dir (doc-view-file-name-to-directory-name doc))
         (png-file (concat dir "/" "page-%d.png")))
    (when (file-exists-p dir)
      (dired-delete-file dir 'always))
    (make-directory dir t)
    ;; DVI files have to be converted to PDF before GhostScript can process it.
    (when (string= (file-name-extension doc) "dvi")
      (message "DocView: converting DVI to PDF.")
      (setq real-doc (concat (file-name-as-directory dir) "doc.pdf"))
      (process-file-shell-command (concat doc-view-dvipdfm-program " -o "
					  real-doc " " doc))
      (message "DocView: conversion to PDF done."))
    (message "DocView: starting conversion to PNG now!")
    (apply 'start-process
           (append (list "doc-view-convert" "*doc-view conversion output*"
                         doc-view-ghostscript-program)
                   doc-view-ghostscript-options
                   (list (concat "-sOutputFile=" png-file))
		   (list real-doc)))))

(define-derived-mode doc-view-mode nil "DocView"
  "Major mode in DocView buffers.

\\{doc-view-mode-map}"
  :group 'doc-view
  (setq buffer-read-only t)
  (make-local-variable 'doc-view-current-files)
  (make-local-variable 'doc-view-current-doc)
  (make-local-variable 'doc-view-current-page)
  (make-local-variable 'doc-view-current-converter-process))

(defun doc-view-display-image (file)
  "Display the given png FILE."
  (setq inhibit-read-only t)
  (erase-buffer)
  (let ((image (create-image file 'png nil
                             :margin doc-view-display-margin
			     :pointer 'hand)))
    (insert-image image))
  (insert (propertize
           (format "\nPage %d of %d." doc-view-current-page (length doc-view-current-files))
           'face 'bold))
  (goto-char (point-min))
  (forward-char)
  (setq inhibit-read-only nil))

(defun doc-view-sort (a b)
  "Return non-nil if A should be sorted before B.
Predicate for sorting `doc-view-current-files'."
  (if (< (length a) (length b))
      t
    (if (> (length a) (length b))
        nil
      (string< a b))))

(defun doc-view-display (doc)
  "Start viewing the document DOC."
  (let ((dir (doc-view-file-name-to-directory-name doc)))
    (switch-to-buffer (format "*DocView: %s*" doc))
    (setq doc-view-current-files
          (sort (directory-files dir t "page-[0-9]+\\.png" t)
		'doc-view-sort))
    (doc-view-goto-page 1)))

(defun doc-view-buffer-message ()
  (setq inhibit-read-only t)
  (erase-buffer)
  (insert (propertize "Welcome to DocView!" 'face 'bold)
	  "\n"
	  "
If you  see this buffer  it means that  the document you  want to
view gets converted  to PNG now.  This may take  some time if the
document is large.

For now these keys are useful:

    `q' : Bury   this   buffer.   It'll pop  up  again  when  the
          conversion finished.
    `k' : Kill the conversion process and this buffer.\n")
  (setq inhibit-read-only nil))

(defun doc-view (no-cache)
  "Query for a document, convert it to png and start viewing it.
If this file is still in the cache, don't convert and use the
existing page files.  With prefix arg NO-CACHE, don't use the
cached files and convert anew."
  (interactive "P")
  (if (not (and (image-type-available-p 'png)
                (display-images-p)))
      (message "DocView: your emacs or display doesn't support png images.")
    (let* ((doc (expand-file-name (read-file-name "File: " nil nil t)))
	   (buffer (get-buffer-create (format "*DocView: %s*" doc)))
	   (dir (doc-view-file-name-to-directory-name doc)))
      (switch-to-buffer buffer)
      (doc-view-buffer-message)
      (doc-view-mode)
      (setq doc-view-current-doc doc)
      (if (and (file-exists-p dir)
	       (not no-cache))
	  (progn
	    (message "DocView: using cached files!")
	    (doc-view-display doc-view-current-doc))
	(setq doc-view-current-converter-process (doc-view-convert-doc doc-view-current-doc))
	(process-put doc-view-current-converter-process 'buffer (current-buffer))
	(set-process-sentinel doc-view-current-converter-process
			      '(lambda (proc event)
				 (set-buffer (process-get proc 'buffer))
				 (setq doc-view-current-converter-process nil)
				 (when (string-match "finished" event)
				   (message "DocView: finished conversion!")
				   (doc-view-display doc-view-current-doc))))))))

(defun doc-view-clear-cache ()
  "Delete the whole cache (`doc-view-cache-directory')."
  (interactive)
  (dired-delete-file doc-view-cache-directory 'always)
  (make-directory doc-view-cache-directory))

(defun doc-view-dired-cache ()
  "Open `dired' in `doc-view-cache-directory'.."
  (interactive)
  (dired doc-view-cache-directory))

(provide 'doc-view)

;;; doc-view.el ends here
